import { BehaviorSubject, Observable, Subject } from "rxjs"
import { Bus, FisAppActor, ReceiverProfile, TransmissionMessage, TransmissionProfile, TransmitterProfile } from "./transport.interface"
import { ConnectionAdapter } from "../connector/connector.base"


export type TYPE = {
    adapterProfile: AdapterProfile,
    transmitterProfile: TransmitterProfile,
    receiverProfile: ReceiverProfile,
    requestResponseProfile: TransmissionProfile
}

export interface AdapterProfile {
    id: string,
    adapter: ConnectionAdapter,
    connectionState: Subject<ConnectionState>
}


export interface ConnectionManager {
    transportService: TransportService
    // to get notified on what's going on in Transmission Manager
    transportEvent: Subject<TransportEvent>
    // list of connection
    adapterSet: AdapterSet[]
    // Called by transmission manager to have an instance of these adapters
    getAdapterSetById(adapterSetid: string): AdapterSet
}

export interface ConnectionAdaptorBase {
    connector: TransportService // this one will refer to the actual tranpsort service like websocket and so on
    connectorProfile: AdapterProfile
    connectionStateBus: Subject<ConnectionState>
    adaptorTransmissionRole: AdaptorTransmissionRole
    event: Subject<TransportEvent>

    subscribeConnectionState(): Observable<ConnectionState>
    publishConnectionState(): void
    connect(): void
    disconnect(): void
    getMessageBus(bus: Bus): Observable<any>
    getInfo(): AdapterProfile
    setAdapterProfile(id: string): void
}


export interface TransmitterConnectionAdapter extends ConnectionAdaptorBase {
    emit(message: TransmissionMessage): void
    emitStream(message: TransmissionMessage): void
}

export interface ReceiverConnectionAdapter extends ConnectionAdaptorBase {
    subscribeMessages(messageFilter: any): Observable<any>
}

export interface RequestResponseConnectionAdapter extends TransmitterConnectionAdapter, ReceiverConnectionAdapter {
    send(message: TransmissionMessage): Observable<TransportMessage>
}

export type ConnectionState = 'ONLINE' | 'OFFLINE'
export enum AdaptorTransmissionRole {
    Transmitter,
    Receiver,
    RequestResponse
}

export enum Transport {
    Websocket = `Websocket`,
    Grpc = `Grpc`,
    Http = `Http`
}

// TO be used for transmission at the trasport level
export interface TransportMessage {
    id: string,
    dateCreated: Date,
    transport: Transport,
    target?: string,
    payload: any
}

export interface TransportEvent {
    id: string,
    event: 'Server Started' | 'New Client' | 'Client Disconnected' | 'Client Reconnected' | `Server Disconnected` | 'New Message' | `Notification` | `New Server` | `New Transport` | 'New Adapter',
    data: any
}

export interface TransportService {
    getInfo(): Transport
    emit(message: TransportMessage): void
    subscribe(): Observable<TransportEvent> //all messages and whatever event will go through this, easier to implemnet across different transport protocol
}


export interface Info {
    transport: Transport
}


export interface AdapterSet {
    id: string,
    dateCreated: Date,
    transmitterAdapter: TransmitterConnectionAdapter,
    receiverAdapter: ReceiverConnectionAdapter,
    requestResponsAdapter: RequestResponseConnectionAdapter
}

export interface ClientObject {
    id: string,
    dateCreated: Date,
    connectionState: BehaviorSubject<ConnectionState>
}
